# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import sys
from os.path import basename, dirname, splitext
from metagenomix._inputs import (sample_inputs, group_inputs, genome_key,
                                 genome_out_dir)
from metagenomix._io_utils import (caller, io_update, to_do, tech_specificity,
                                   not_paired, status_update)


def predict_cmd(
        self,
        fasta: str,
        out: str,
        typ: str
) -> str:
    """Collect deeparg predict commands.

    Parameters
    ----------
    self : Commands class instance
        .soft.params
            Parameters
    fasta : str
        Path to the input fasta file
    out : str
        Paths to the output file's prefix
    typ : str
        Type of input data: 'nucl' (DNA) or 'prot' (protein)

    Returns
    -------
    cmd : str
        deeparg predict commands
    """
    cmd = 'deeparg predict'
    cmd += ' --input-file %s' % fasta
    cmd += ' --output-file %s' % out
    cmd += ' --data-path %s' % self.soft.params['db_dir']
    cmd += ' --type %s' % typ
    for param in [
        'min_prob', 'arg_alignment_overlap', 'arg_alignment_evalue',
        'arg_alignment_identity', 'arg_num_alignments_per_entry',
        'model_version'
    ]:
        cmd += ' --%s %s' % (param.replace('_', '-'), self.soft.params[param])
    if self.soft.prev == 'plass':
        cmd += ' --model SS'
    else:
        cmd += ' --model LS'
    return cmd


def predict_inputs(
        self,
        fasta: str
) -> dict:
    """Get input files per sequence type.

    Parameters
    ----------
    self : Commands class instance
        .soft.prev : str
            Previous software in the pipeline
    fasta : str
        Path to the input file or folder

    Returns
    -------
    typ_seqs : dict
        Paths to the input fasta file per sequence type
    """
    if self.soft.prev == 'prodigal':
        typ_seqs = {'prot': '%s/protein.translations.fasta' % fasta,
                    'nucl': '%s/nucleotide.sequences.fasta' % fasta}
    elif self.soft.prev == 'plass':
        typ_seqs = {'prot': '%s/prot_contigs.fasta' % dirname(fasta),
                    'nucl': '%s/nucl_contigs.fasta' % dirname(fasta)}
    else:
        sys.exit('[%s] Only avail after prodigal or plass' % self.soft.name)
    return typ_seqs


def get_predict(
        self,
        fastas: dict,
        tech: str,
        sam_group: str
) -> None:
    """

    Parameters
    ----------
    self : Commands class instance
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
    fastas : dict
        Paths to the input fasta files per genome/MAG
    tech : str
        Technology: 'illumina', 'pacbio', or 'nanopore'
    sam_group : str
        Sample name or name of a co-assembly pool's group
    """
    for genome, fasta in fastas.items():

        out_dir = genome_out_dir(self, tech, fasta[0], sam_group, genome)
        self.outputs['dirs'].append(out_dir)

        typ_seqs = predict_inputs(self, fasta[0])
        for typ, seq in typ_seqs.items():

            key = genome_key(tech, sam_group, genome)
            status_update(self, tech, [seq], group=sam_group, genome=genome)

            base = splitext(basename(seq))[0]
            prefix = '%s/%s' % (out_dir, base)
            arg = '%s.mapping.ARG' % prefix
            pot_arg = '%s.potential.ARG' % prefix
            outs = [arg, pot_arg]
            self.outputs['outs'].setdefault((tech, sam_group), []).append(outs)

            # check if tool already run (or if --force) to allow getting command
            if self.config.force or to_do(arg):
                # collect the command line
                cmd = predict_cmd(self, seq, prefix, typ)
                # add is to the 'cmds'
                self.outputs['cmds'].setdefault(key, []).append(cmd)
                io_update(self, i_f=fasta[0], o_d=out_dir, key=key)
                self.soft.add_status(
                    tech, self.sam_pool, 1, group=sam_group, genome=genome)
            else:
                self.soft.add_status(
                    tech, self.sam_pool, 0, group=sam_group, genome=genome)


def predict(self) -> None:
    """Run deeparg predict module.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for metaWRAP
        .pool : str
            Pool name
        .pools : dict
            Co-assembly pools and sample per group
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
    """
    if self.sam_pool in self.pools:
        for (tech, group), inputs in self.inputs[self.sam_pool].items():
            fastas = group_inputs(self, inputs)
            get_predict(self, fastas, tech, group)

    elif self.soft.prev == 'drep':
        for (tech, bin_algo), inputs in self.inputs[''].items():
            fastas = group_inputs(self, inputs)
            get_predict(self, fastas, tech, bin_algo)
    else:
        tech_fastas = sample_inputs(self)
        for tech, fastas in tech_fastas.items():
            get_predict(self, fastas, tech, self.sam_pool)


def short_cmd(
        self,
        fastqs: list,
        prefix: str
) -> str:
    """Collect deeparg short_reads_pipeline commands.

    Parameters
    ----------
    self : Commands class instance
        .soft.params
            Parameters
    fastqs : list
        Paths to the input files
    prefix : str
        Paths to the output file's prefix

    Returns
    -------
    cmd : str
        deeparg short_reads_pipeline commands
    """
    cmd = 'deeparg short_reads_pipeline'
    cmd += ' --forward_pe_file %s --reverse_pe_file %s' % tuple(fastqs)
    cmd += ' --output_file %s' % prefix
    for param in [
        'deeparg_data_path', 'deeparg_identity', 'deeparg_probability',
        'deeparg_evalue', 'gene_coverage', 'bowtie_16s_identity'
    ]:
        cmd += ' --%s %s' % (param.replace('_', '-'), self.soft.params[param])
    return cmd


def short(self) -> None:
    """Run deeparg short_reads_pipeline module.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for metaWRAP
        .sam_pool : str
            Sample of co-assembly group name.
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
    """
    # iterate over the inputs
    for (tech, sam), fastqs in self.inputs[self.sam_pool].items():
        if tech_specificity(self, fastqs, tech, sam, ['illumina']):
            continue
        if not_paired(self, tech, sam, fastqs):
            continue
        status_update(self, tech, fastqs)

        # make the output directory
        out = '%s/%s/%s' % (self.dir, tech, self.sam_pool)
        self.outputs['dirs'].append(out)

        # get the expected names of some of the ouptuts:
        # - those you want to collect in 'outs': will be used as future inputs
        # - at least one: will help to know whether the software already run
        prefix = out + '/' + self.sam_pool
        arg = '%s.mapping.ARG' % prefix
        pot_arg = '%s.potential.ARG' % prefix
        outs = [arg, pot_arg]
        self.outputs['outs'].setdefault((tech, self.sam_pool), []).extend(outs)

        # check if the tool already run (or if --force) to allow getting command
        if self.config.force or to_do(arg):
            # collect the commmand line
            cmd = short_cmd(self, fastqs, prefix)
            # add is to the 'cmds'
            self.outputs['cmds'][(tech,)] = [cmd]
            io_update(self, i_f=fastqs, o_d=out, key=tech)
            self.soft.add_status(tech, sam, 1)
        else:
            self.soft.add_status(tech, sam, 0)


def deeparg(self) -> None:
    """A deep learning based approach to predict Antibiotic Resistance Genes
    (ARGs) from metagenomes. It provides two models,deepARG-SS and deepARG-LS.

    References
    ----------
    Arango-Argoty, Gustavo, et al. "DeepARG: a deep learning approach for
    predicting antibiotic resistance genes from metagenomic data." Microbiome
    6.1 (2018): 1-15.

    Notes
    -----
    BitBucket   : https://bitbucket.org/gusphdproj/deeparg-ss/src/master/
    Website     : https://bench.cs.vt.edu/deeparg
    Docs        : https://readthedocs.org/projects/deeparg/
    Paper       : https://doi.org/10.1186/s40168-018-0401-z

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for metaWRAP
        .sam_pool : str
            Sample of co-assembly group name.
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
    """
    # This function splits the name of the software and calls as function
    # the last underscore-separated field (which is in this module)
    module_call = caller(self, __name__)
    module_call(self)


def mmarc_cmd(
        self,
        tech: str,
        inputs: list,
        out: str,
        genome: str,
        level: str
) -> str:
    """Collect mmarc (meta-marc) command.

    Parameters
    ----------
    self : Commands class instance
        .soft.params
            Parameters
    tech : str
        Technology: 'illumina', 'pacbio', or 'nanopore'
    inputs : list
        Paths to the input files
    out : str
        Paths to the output folder
    genome : str
        MAGs/Genomes folder name or empty string (for assembly contigs)
    level : str
        Model level

    Returns
    -------
    cmd : str
        mmarc (meta-marc) command
    """
    cmd = '\n./mmarc'
    if genome:
        cmd += ' --input %s' % inputs[0]
    else:
        if len(inputs) == 1:
            cmd += ' --input %s' % inputs[0]
        if len(inputs) == 2:
            cmd += ' --i1 %s --i2 %s' % tuple(inputs)
        if tech == 'illumina':
            cmd += ' --dedup'
        cmd += ' --multicorrect'
    cmd += ' --threads %s' % self.soft.params['cpus']
    cmd += ' --output %s' % out
    cmd += ' --filename output'
    cmd += ' --skewness %s/skewness.txt' % out
    cmd += ' --graphs %s/graphs' % out
    cmd += ' --level %s' % level

    for boolean in ['dedup', 'multicorrect']:
        if self.soft.params[boolean]:
            cmd += ' --%s' % boolean

    for param in ['coverage', 'evalue', 'kmer']:
        cmd += ' --%s %s' % (param.replace('_', '-'), self.soft.params[param])

    return cmd


def get_metamarc(
        self,
        tech: str,
        fastas_folders: dict,
        sam_group: str
) -> None:
    """

    Parameters
    ----------
    self : Commands class instance
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .soft.status
            Current status of the pipeline in terms of available outputs
        .config
            Configurations
    tech : str
        Technology: 'illumina', 'pacbio', or 'nanopore'
    fastas_folders : dict
        Paths to the input fasta files per genome/MAG
    sam_group : str
        Name of the current sample or co-assembly group
    """
    for genome, inputs in fastas_folders.items():

        fasta = inputs[0]
        out_dir = genome_out_dir(self, tech, fasta, sam_group, genome)
        self.outputs['outs'].setdefault((tech, sam_group), []).append(out_dir)
        status_update(
            self, tech, inputs, self.sam_pool, group=sam_group, genome=genome)

        outs, cmd = [], ''
        key = genome_key(tech, sam_group, genome)
        for level in self.soft.params['level']:
            out = '%s/model_level_%s' % (out_dir, level)
            self.outputs['dirs'].append(out_dir)
            outs.append(out)
            skewness = '%s/skewness.txt' % out
            # check if tool already run (or if --force) to allow getting command
            if self.config.force or to_do(skewness):
                # collect the command line
                cmd += mmarc_cmd(self, tech, inputs, out, genome, level)

        if cmd:
            path = self.soft.params['path']
            full_cmd = 'cd %s\n' % out_dir
            full_cmd += 'cp -r %s %s/.\n' % (path, out_dir)
            full_cmd += 'cp -r %s/../hmmer-3.1b2 %s/.\n' % (path, out_dir)
            full_cmd += 'cp -r %s/../hmmer-3.1b2 %s/.\n' % (path, out_dir)
            full_cmd += 'cd bin\n'
            full_cmd += cmd
            full_cmd += '\nrm -rf bin\n'
            full_cmd += 'rm -rf hmmer-3.1b2\n'
            # add is to the 'cmds'
            self.outputs['cmds'].setdefault(key, []).append(full_cmd)
            io_update(self, i_f=inputs, o_d=outs, key=key)
            self.soft.add_status(
                tech, self.sam_pool, 1, group=sam_group, genome=genome)
        else:
            self.soft.add_status(
                tech, self.sam_pool, 0, group=sam_group, genome=genome)


def metamarc(self) -> None:
    """Meta-MARC is a set of profile Hidden Markov Models developed for the
    purpose of screening and profiling resistance genes in DNA-based
    metagenomic data. This tool was developed for the characterization of
    various resistance classes, mechanisms, and gene/operon groups from raw
    sequencing data much in the way that microbiome softwares profile the
    bacterial taxonomy of metagenomic samples. Meta-MARC is not intended to
    be used as a final annotation or all-in-one tool; this software simply
    offers the user another view of complex metagenomic data. Users are
    encouraged to perform further statistical testing on their data to
    validate results obtained from Meta-MARC.

    References
    ----------
    Lakin, Steven M., et al. "Hierarchical Hidden Markov models enable
    accurate and diverse detection of antimicrobial resistance sequences."
    Communications biology 2.1 (2019): 1-11.

    Notes
    -----
    GitHub  : https://github.com/lakinsm/meta-marc
    Paper   : https://doi.org/10.1038/s42003-019-0545-9

    Parameters
    ----------
    self : Commands class instance
        .name : str
            Name of the current software in the pipeline
        .dir : str
            Path to pipeline output folder for metaWRAP
        .sam_pool : str
            Sample of co-assembly group name
        .pools : dict
            Co-assembly pools and sample per group
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.status
            Current status of the pipeline in terms of available outputs
        .soft.params
            Parameters
        .config
            Configurations
    """
    if self.sam_pool in self.pools:
        for (tech, group), inputs in self.inputs[self.sam_pool].items():
            fastas_folders = group_inputs(self, inputs)
            get_metamarc(self, tech, fastas_folders, group)

    elif self.soft.prev == 'drep':
        for (tech, bin_algo), inputs in self.inputs[''].items():
            folders = group_inputs(self, inputs)
            get_metamarc(self, tech, folders, bin_algo)

    else:
        tech_fastas = sample_inputs(self, raw=True)
        for tech, fastas in tech_fastas.items():
            get_metamarc(self, tech, fastas, self.sam_pool)


def karga_kargva_cmd(
        self,
        merged: str,
        db_path: str
) -> str:
    """Collect KARGA or KARGVA command.

    Parameters
    ----------
    self : Commands class instance
        .name : str
            Name of the current software in the pipeline
        .soft.params
            Parameters
    merged : str
        Path to the input file
    db_path : str
        Path to ARG/MGE fasta file with resistance annotation/mutation in header

    Returns
    -------
    cmd : str
        KARGA or KARGVA command
    """
    cmd = 'java %s' % self.soft.name.upper()
    cmd += ' f:%s' % merged
    cmd += ' d:%s' % db_path
    for param in ['k', 'i', 's']:
        if param == 's' and self.soft.name != 'karga':
            continue
        if self.soft.params[param]:
            cmd += ' %s:%s' % (param, self.soft.params[param])
    cmd += ' -Xmx16GB\n'
    return cmd


def get_karga_kargva(
        self,
        tech: str,
        fastas_folders: dict,
        sam_group: str
) -> None:
    """

    Parameters
    ----------
    self : Commands class instance
        .soft.name : str
            Name of the current software in the pipeline
        .soft.prev : str
            Name of the previous software in the pipeline
        .outputs : dict
            All outputs
        .soft.status
            Current status of the pipeline in terms of available outputs
        .soft.params
            Parameters
        .config
            Configurations
    tech : str
        Technology: 'illumina', 'pacbio', or 'nanopore'
    fastas_folders : dict
        Paths to the input fasta files per genome/MAG
    sam_group : str
        Name of the current sample or co-assembly group
    """
    for genome, inputs in fastas_folders.items():

        fastx = inputs[0]
        out_dir = genome_out_dir(self, tech, fastx, sam_group)
        self.outputs['outs'].setdefault((tech, sam_group), []).append(out_dir)
        status_update(self, tech, inputs, group=sam_group, genome=genome)

        merge_cmd = ''
        if len(inputs) == 1:
            merged = fastx
        elif len(inputs) > 1:
            merged = '%s/%s' % (out_dir, basename(fastx))
            merge_cmd += 'cat %s > %s\n' % (' '.join(inputs), merged)

        outs, cmd = [], ''
        for db, db_path in self.soft.params['databases'].items():
            out = '%s/db_%s' % (out_dir, db)
            self.outputs['dirs'].append(out_dir)
            outs.append(out)
            bas = splitext(basename(merged))[0]
            fp = '%s/%s_%s_mappedReads.csv' % (out, self.soft.name.upper(), bas)
            # check if tool already run (or if --force) to allow getting command
            if self.config.force or to_do(fp):
                # collect the command line
                cmd += karga_kargva_cmd(self, merged, db_path)

        if cmd:
            if self.soft.name == 'karga':
                ps = ['openjdk-8', 'AMRGene.class', 'KARGA$1.class',
                      'KARGA.class', 'KARGA.java']
            elif self.soft.name == 'kargva':
                ps = ['AMRVariantGene.class', 'KARGVA$1.class', 'KARGVA.class',
                      'KARGVA.java', 'kargva_db_v5.fasta']
            path = self.soft.params['path']
            full_cmd = 'cd %s\n' % out_dir
            for p in ps:
                full_cmd += 'cp -r %s/%s %s/.\n' % (path, p, out_dir)
            full_cmd += merge_cmd + cmd
            for p in ps:
                full_cmd += 'rm -rf %s\n' % p

            # add is to the 'cmds'
            key = genome_key(tech, sam_group)
            self.outputs['cmds'].setdefault(key, []).append(full_cmd)
            io_update(self, i_f=inputs, o_d=outs, key=key)
            self.soft.add_status(
                tech, self.sam_pool, 1, group=sam_group, genome=genome)
        else:
            self.soft.add_status(
                tech, self.sam_pool, 0, group=sam_group, genome=genome)


def karga_kargva(self) -> None:
    """Get the input for KARGA or KARGVA

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for metaWRAP
        .sam_pool : str
            Sample of co-assembly group name
        .pools : dict
            Co-assembly pools and sample per group
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.name : str
            Name of the current software in the pipeline
        .soft.prev : str
            Name of the previous software in the pipeline
        .soft.status
            Current status of the pipeline in terms of available outputs
        .soft.params
            Parameters
        .config
            Configurations
    """
    if not self.sam_pool or self.sam_pool in self.pools:
        sys.exit('[%s] Only run on non-assembled reads' % self.soft.name)
    else:
        tech_fastas = sample_inputs(self, raw=True)
        for tech, fastas in tech_fastas.items():
            get_karga_kargva(self, tech, fastas, self.sam_pool)


def karga(self) -> None:
    """K-mer-based antibiotic gene resistance analyzer, a multi-platform Java
    toolkit for identifying ARGs from metagenomic short read data. KARGA does
    not perform alignment; it uses an efficient double-lookup strategy,
    statistical filtering on false positives, and provides individual read
    classification as well as covering of the database resistome. On simulated
    data, KARGA’s antibiotic resistance class recall is 99.89% for error/
    mutation rates within 10%, and of 83.37% for error/mutation rates between
    10% and 25%, while it is 99.92% on ARGs with rearrangements. On empirical
    data, KARGA provides higher hit score (≥1.5-fold) than AMRPlusPlus, DeepARG,
    and MetaMARC. KARGA has also faster runtimes than all other softwares (2x faster
    than AMRPlusPlus, 7x than DeepARG, and over 100x than MetaMARC).

    References
    ----------
    M. Prosperi and S. Marini, "KARGA: Multi-platform Toolkit for k-mer-based
    Antibiotic Resistance Gene Analysis of High-throughput Sequencing Data,"
    2021 IEEE EMBS International Conference on Biomedical and Health Informatics
    (BHI), 2021, pp. 1-4, doi: 10.1109/BHI50953.2021.9508479.

    Notes
    -----
    GitHub  : https://github.com/DataIntellSystLab/KARGA
    Paper   : https://ieeexplore.ieee.org/document/9508479

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for metaWRAP
        .sam_pool : str
            Sample of co-assembly group name
        .pools : dict
            Co-assembly pools and sample per group
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.name : str
            Name of the current software in the pipeline
        .soft.prev : str
            Name of the previous software in the pipeline
        .soft.status
            Current status of the pipeline in terms of available outputs
        .soft.params
            Parameters
        .config
            Configurations
    """
    karga_kargva(self)


def kargva(self) -> None:
    """K-mer-based Antibiotic Resistance Gene Variant Analyzer (KARGVA).
    KARGVA is a Multi-platform Toolkit for Identification of Antibiotic
    Resistance from Sequencing Data Conferred by Point Mutations in
    Bacterial Genes.

    References
    ----------
    M. Prosperi and S. Marini, "KARGA: Multi-platform Toolkit for k-mer-based
    Antibiotic Resistance Gene Analysis of High-throughput Sequencing Data,"
    2021 IEEE EMBS International Conference on Biomedical and Health Informatics
    (BHI), 2021, pp. 1-4, doi: 10.1109/BHI50953.2021.9508479.

    Notes
    -----
    GitHub  : https://github.com/DataIntellSystLab/KARGVA
    Paper   : https://ieeexplore.ieee.org/document/9508479

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for metaWRAP
        .sam_pool : str
            Sample of co-assembly group name
        .pools : dict
            Co-assembly pools and sample per group
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.name : str
            Name of the current software in the pipeline
        .soft.prev : str
            Name of the previous software in the pipeline
        .soft.status
            Current status of the pipeline in terms of available outputs
        .soft.params
            Parameters
        .config
            Configurations
    """
    karga_kargva(self)


def abricate_cmd(
        self,
        tech: str,
        inputs: list,
        out_d: str,
) -> str:
    """Collect abricate command.

    Parameters
    ----------
    self : Commands class instance
        .soft.params
            Parameters
    inputs : list
        Paths to the input files
    out_d : str
        Paths to the output folder

    Returns
    -------
    cmd : str
        abricate command
    """
    cmd, seqtk_cmd = '', ''
    for db in self.soft.params['databases']:

        cmd += '\nabricate'

        if tech in ['nanopore', 'pacbio']:
            outs = []
            for idx, inp in enumerate(inputs):
                out = '%s/%s.fa' % (out_d, idx)
                outs.append(out)
                seqtk_cmd += 'seqtk seq -A %s > %s/%s\n' % (inp, out_d, out)
            out = '%.fasta' % splitext(basename(inputs[0]))[0]
            seqtk_cmd += 'cat %s > %s/%s\n' % (' '.join(outs), out_d, out)
        else:
            out = inputs[0]

        cmd += ' %s' % out
        cmd += ' --threads %s' % self.soft.params['cpus']
        cmd += ' --db %s' % db
        for boolean in ['setupdb', 'noheader', 'csv', 'nopath']:
            if self.soft.params[boolean]:
                cmd += ' --%s' % boolean
        for param in ['minid', 'mincov']:
            cmd += ' --%s %s' % (param, self.soft.params[param])
        cmd += ' > %s/%s.txt' % (out_d, db)

    if seqtk_cmd:
        cmd = seqtk_cmd + cmd

    return cmd


def get_abricate(
        self,
        tech: str,
        fastas_folders: dict,
        sam_group: str
) -> None:
    """

    Parameters
    ----------
    self : Commands class instance
        .outputs : dict
            All outputs
        .soft.status
            Current status of the pipeline in terms of available outputs
        .soft.params
            Parameters
        .config
            Configurations
    tech : str
        Technology: 'illumina', 'pacbio', or 'nanopore'
    fastas_folders : dict
        Paths to the input fasta files per genome/MAG
    sam_group : str
        Name of the current sample or co-assembly group
    """
    for genome, inputs in fastas_folders.items():

        fasta = inputs[0]
        out_d = genome_out_dir(self, tech, fasta, sam_group, genome)
        self.outputs['dirs'].append(out_d)
        self.outputs['outs'].setdefault((tech, sam_group), []).append(out_d)

        status_update(self, tech, [fasta], group=sam_group, genome=genome)

        outs = ['%s/%s.txt' % (out_d, x) for x in self.soft.params['databases']]
        # check if tool already run (or if --force) to allow getting command
        if self.config.force or sum([to_do(x) for x in outs]):
            # collect the command line
            cmd = abricate_cmd(self, tech, inputs, out_d)
            # add is to the 'cmds'
            key = genome_key(tech, sam_group, genome)
            self.outputs['cmds'].setdefault(key, []).append(cmd)
            io_update(self, i_f=inputs, o_d=out_d, key=key)
            self.soft.add_status(
                tech, self.sam_pool, 1, group=sam_group, genome=genome)
        else:
            self.soft.add_status(
                tech, self.sam_pool, 0, group=sam_group, genome=genome)


def abricate(self) -> None:
    """Mass screening of contigs for antimicrobial resistance or virulence
    genes. It comes bundled with multiple databases: NCBI, CARD, ARG-ANNOT,
    Resfinder, MEGARES, EcOH, PlasmidFinder, Ecoli_VF and VFDB.

    Notes
    -----
    GitHub  : https://github.com/tseemann/abricate

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for metaWRAP
        .sam_pool : str
            Sample of co-assembly group name
        .pools : dict
            Co-assembly pools and sample per group
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.name : str
            Name of the current software in the pipeline
        .soft.prev : str
            Name of the previous software in the pipeline
        .soft.status
            Current status of the pipeline in terms of available outputs
        .soft.params
            Parameters
        .config
            Configurations
    """
    if self.sam_pool in self.pools:
        for (tech, group), inputs in self.inputs[self.sam_pool].items():
            fastas_folders = group_inputs(self, inputs)
            get_abricate(self, tech, fastas_folders, group)

    elif self.soft.prev == 'drep':
        for (tech, bin_algo), inputs in self.inputs[''].items():
            folders = group_inputs(self, inputs)
            get_abricate(self, tech, folders, bin_algo)

    else:
        tech_fastas = sample_inputs(self, ['nanopore', 'pacbio'], raw=True)
        for tech, fastas in tech_fastas.items():
            get_abricate(self, tech, fastas, self.sam_pool)


def amrplusplus2_cmd(
        self,
        fastqs: list,
        out_dir: str,
) -> str:
    """Collect amrplusplus2 command.

    Parameters
    ----------
    self : Commands class instance
        .soft.params
            Parameters
    fastqs : list
        Paths to the input files
    out_dir : str
        Paths to the output folder

    Returns
    -------
    cmd : str
        amrplusplus2 command
    """
    cmd = 'mkdir %s/copy\n' % out_dir
    cmd += 'cp -r %s/* %s/copy/.\n' % (self.soft.params['path'], out_dir)
    cmd += 'cd %s/copy\n' % out_dir

    cmd += 'nextflow run main_AmrPlusPlus_v2.nf'
    # input argument
    if len(fastqs) > 1:
        cmd += ' --reads "%s"' % fastqs[0].replace('R1', '{R1,R2}')
    else:
        cmd += ' --reads %s' % fastqs[0]

    # optional arguments - "valued"
    for param in ['adapters', 'fqc_adapters', 'host_index', 'host',
                 'kraken_db', 'amr_index', 'amr', 'annotation',
                 'snp_annotation', 'snp_confirmation']:
        if self.soft.params[param]:
            cmd += ' --%s "%s"' % (param, self.soft.params[param])

    for param in ['leading', 'trailing', 'slidingwindow', 'minlen',
                  'threshold', 'min', 'max', 'skip', 'samples']:
        cmd += ' --%s %s' % (param, self.soft.params[param])
    cmd += ' --threads %s' % self.soft.params['cpus']
    cmd += ' --output "%s"' % out_dir
    cmd += ' -work "%s/work_dir"\n' % out_dir
    cmd += 'cd %s\n' % out_dir
    cmd += 'rm -rf copy\n'
    cmd += 'rm -rf work_dir\n'
    cmd += 'rm -rf RunQC\n'
    cmd += 'rm -rf BuildHostIndex\n'
    cmd += 'rm -rf AlignReadsToHost\n'
    cmd += 'rm -rf NonHostReads\n'
    cmd += 'rm -rf RemoveHostDNA\n'
    cmd += 'rm -rf RunKraken\n'
    cmd += 'rm -rf KrakenResults\n'
    cmd += 'rm -rf FilteredKrakenResults\n'
    return cmd


def get_amrplusplus2(
        self,
        tech: str,
        fastqs: list,
) -> None:
    """

    Parameters
    ----------
    self : Commands class instance
        .outputs : dict
            All outputs
        .soft.status
            Current status of the pipeline in terms of available outputs
        .soft.params
            Parameters
        .config
            Configurations
    tech : str
        Technology: 'illumina', 'pacbio', or 'nanopore'
    fastqs : list
        Paths to the input fastqs files
    """
    out_dir = genome_out_dir(self, tech, fastqs[0], self.sam_pool)
    self.outputs['dirs'].append(out_dir)
    self.outputs['outs'].setdefault((tech, self.sam_pool), []).append(out_dir)

    status_update(self, tech, fastqs)

    # check if tool already run (or if --force) to allow getting command
    out = '%s/ResistomeResults/AMR_analytic_matrix.csv' % out_dir
    if self.config.force or to_do(out):
        # collect the command line
        cmd = amrplusplus2_cmd(self, fastqs, out_dir)
        # add is to the 'cmds'
        key = genome_key(tech, self.sam_pool)
        self.outputs['cmds'].setdefault(key, []).append(cmd)
        io_update(self, i_f=fastqs, o_d=out_dir, key=key)
        self.soft.add_status(tech, self.sam_pool, 1)
    else:
        self.soft.add_status(tech, self.sam_pool, 0)


def amrplusplus2(self) -> None:
    """AmrPlusPlus is an easy to use app that identifies and characterizes
    resistance genes within sequence data.

    References
    ----------
    Lakin, Steven M., et al. "MEGARes: an antimicrobial resistance database
    for high throughput sequencing." Nucleic acids research 45.D1 (2017):
    D574-D580.

    Notes
    -----
    GitHub  : https://github.com/meglab-metagenomics/amrplusplus_v2
    Docs    : http://megares.meglab.org/amrplusplus/latest/html/index.html
    Paper   : https://doi.org/10.1093/nar/gkw1009

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for metaWRAP
        .sam_pool : str
            Sample of co-assembly group name
        .pools : dict
            Co-assembly pools and sample per group
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.name : str
            Name of the current software in the pipeline
        .soft.prev : str
            Name of the previous software in the pipeline
        .soft.status
            Current status of the pipeline in terms of available outputs
        .soft.params
            Parameters
        .config
            Configurations
    """
    if not self.sam_pool or self.sam_pool in self.pools:
        sys.exit('[%s] Only run on non-assembled reads' % self.soft.name)
    else:
        tech_fastqs = sample_inputs(self, raw=True)
        for tech, fastqs in tech_fastqs.items():
            get_amrplusplus2(self, tech, fastqs[''])


def amrfinderplus(self) -> None:
    """This software and the accompanying database are designed to find
    acquired antimicrobial resistance genes and some point mutations in
    protein or assembled nucleotide sequences. We have also added "plus"
    stress, head, and biocide resistance as well as some virulence factors
    and E. coli antigens.

    References
    ----------
    Feldgarden, M., Brover, V., Gonzalez-Escalona, N., Frye, J.G., Haendiges,
    J., Haft, D.H., Hoffmann, M., Pettengill, J.B., Prasad, A.B., Tillman,
    G.E. and Tyson, G.H., 2021. AMRFinderPlus and the Reference Gene Catalog
    facilitate examination of the genomic links among antimicrobial
    resistance, stress response, and virulence. Scientific reports, 11(1),
    pp.1-9.

    Notes
    -----
    GitHub  : https://github.com/ncbi/amr
    Docs    : https://www.ncbi.nlm.nih.gov/pathogens/antimicrobial-resistance/AMRFinder/
    Wiki    : https://github.com/ncbi/amr/wiki
    Paper   : https://doi.org/10.1038/s41598-021-91456-0

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for AMRFinderPlus
        .sam_pool : str
            Sample of co-assembly group name
        .pools : dict
            Co-assembly pools and sample per group
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.name : str
            Name of the current software in the pipeline
        .soft.prev : str
            Name of the previous software in the pipeline
        .soft.status
            Current status of the pipeline in terms of available outputs
        .soft.params
            Parameters
        .config
            Configurations
    """
    # # iterate over the inputs
    #
    #     # make the output directory
    #
    #     # get the expected names of some of the ouptuts:
    #     # - those you want to collect in 'outs' because they will be future inpt
    #     # - at least one that will help knowing whether the software already run
    #
    #     # check if the tool already run (or if --force) to allow getting command
    #     if self.config.force or :
    #         # collect the command line
    #         # add is to the 'cmds'
    #     else:
    #
    pass


def argsoap(self) -> None:
    """.

    References
    ----------

    Notes
    -----
    GitHub  :
    Docs    :
    Paper   :

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for ARGs-OAP
        .sam_pool : str
            Sample of co-assembly group name
        .pools : dict
            Co-assembly pools and sample per group
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.name : str
            Name of the current software in the pipeline
        .soft.prev : str
            Name of the previous software in the pipeline
        .soft.status
            Current status of the pipeline in terms of available outputs
        .soft.params
            Parameters
        .config
            Configurations
    """
    pass



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import glob
from os.path import dirname
from metagenomix.core.parameters import tech_params
from metagenomix._io_utils import (io_update, tech_specificity,
                                   to_do, status_update)


def get_read_count(
        self,
        tech: str,
        sam: str,
) -> str:
    """Get the number of reads in the current sample,
    or an empty string if the counting did not happen yet.

    Parameters
    ----------
    self : Commands class instance
        .softs : dict of software classes
    tech : str
        Technology: 'illumina', 'pacbio', or 'nanopore'
    sam: str
        Sample name

    Returns
    -------
    reads : str
        Number of reads
    """
    reads = ''
    if 'count' in self.softs:
        reads_fps = self.softs['count'].outputs.get((sam, tech), [])
        if not to_do(reads_fps):
            reads_fp = reads_fps.replace('${SCRATCH_FOLDER}', '')
            with open(reads_fp) as f:
                for line in f:
                    reads = line[-1].strip().split(',')[-1]
                    break
        else:
            t = "marker_ab_table"
            if t in tech_params(self, tech)['t']:
                self.soft.messages.add('Run "count" to use read counts for' + t)
    return reads


def analyses(
        self,
        tech: str,
        sam: str,
        bowtie2out: str,
        reads: str
) -> None:
    """Get the taxonomic analyses Metaphlan command.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for humann
        .sam : str
            Sample name
        .outputs : dict
            All outputs
        .soft.params
            Parameters for humann
        .config
            Configurations
    tech : str
        Technology: 'illumina', 'pacbio', or 'nanopore'
    sam : str
        Name of the current sample
    bowtie2out : str
        Input type was a bowtie2 output of the same tool
    reads : str
        Number of reads in the current sample
    """
    params = tech_params(self, tech)
    if 'a' in params:
        tax_levs = ['a']
    else:
        tax_levs = params['tax_lev']
    for t in params['t']:
        out = '%s/%s/%s' % (self.dir, tech, t)
        for tax_lev in tax_levs:
            rad = '%s/%s_t-%s' % (out, sam, tax_lev)
            cmd = 'metaphlan'
            cmd += ' %s' % bowtie2out
            cmd += ' -t %s' % t
            cmd += ' --tax_lev %s' % tax_lev
            cmd += ' --input_type bowtie2out'
            cmd += ' --sample_id %s' % sam
            cmd += ' --sample_id_key sample_name'
            cmd += ' --nproc %s' % params['cpus']
            for param in [
                'perc_nonzero',
                'stat_q',
                'min_cu_len',
                'min_alignment_len',
            ]:
                cmd += ' --%s %s' % (param, params[param])
            for param in [
                'ignore_eukaryotes',
                'ignore_bacteria',
                'ignore_archaea',
                'add_viruses',
                'avoid_disqm'
            ]:
                if params[param]:
                    cmd += ' --%s' % param
            if reads and t == 'marker_ab_table':
                cmd += ' --nreads %s' % reads
            elif t == 'marker_pres_table':
                cmd += ' --pres_th %s' % params['pres_th']
            elif t == 'clade_specific_strain_tracker':
                for strain in self.config.strains:
                    cms = cmd
                    cms += ' --clade %s' % strain.replace(' ', '_')
                    clade_out = '%s_%s.tsv' % (rad, strain.replace(' ', '_'))
                    cms += ' --output_file %s' % clade_out
                    if params['min_ab']:
                        cms += ' --min_ab %s' % params['min_ab']
                    if self.config.force or to_do(clade_out):
                        self.outputs['cmds'].setdefault((tech,), []).append(cmd)
                        io_update(self, o_f=clade_out, key=tech)
                        self.soft.add_status(
                            tech, self.sam_pool, 1, message=t, genome='strain')
                    else:
                        self.soft.add_status(
                            tech, self.sam_pool, 0, message=t, genome='strain')
            else:
                ab_out = '%s.tsv' % rad
                cmd += ' --output_file %s' % ab_out
                if self.config.force or to_do(ab_out):
                    self.outputs['cmds'].setdefault((tech,), []).append(cmd)
                    io_update(self, o_f=ab_out, key=tech)
                    self.soft.add_status(tech, self.sam_pool, 1, message=t)
                else:
                    self.soft.add_status(tech, self.sam_pool, 0, message=t)


def profiling(
        self,
        tech: str,
        sam: str,
        inputs: list,
        bowtie2out: str,
        tmpdir: str
) -> tuple:
    """Get the taxonomic profiling metaphlan command.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for MIDAS
        .sam : str
            Sample name
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .databases
            All databases class instance
        .config
            Configurations
    tech : str
        Technology: 'illumina', 'pacbio', or 'nanopore'
    sam : str
        Name of the current sample
    inputs : list
        Paths for the input fastq files
    bowtie2out : str
        Input type was a bowtie2 output of the same tool
    tmpdir : str
        Path to a temporary directory

    Returns
    -------
    sam_out : str
        Alignment output
    profile_out : str
        Metaphlan taxonomic profile output
    """
    sam_out = '%s/%s/sams/%s.sam.bz2' % (self.dir, tech, sam)
    profile_out = '%s/%s/profiles/%s.tsv' % (self.dir, tech, sam)

    params = tech_params(self, tech)
    if to_do(profile_out):
        io_update(self, o_f=profile_out, key=tech)
        cmd = 'metaphlan'
        if not to_do(bowtie2out):
            io_update(self, i_f=bowtie2out, key=tech)
            cmd += ' %s' % bowtie2out
            cmd += ' --input_type bowtie2out'
        else:
            status_update(self, tech, inputs)
            io_update(self, i_f=inputs, key=tech)
            cmd += ' %s' % ','.join(inputs)
            cmd += ' --input_type fastq'
            cmd += ' --samout %s' % sam_out
            cmd += ' --bt2_ps %s' % params['bt2_ps']
            cmd += ' --min_mapq_val %s' % params['min_mapq_val']
            cmd += ' --read_min_len %s' % params['read_min_len']
            cmd += ' --bowtie2out %s' % bowtie2out
            cmd += ' --bowtie2db %s' % self.databases.paths['metaphlan']
        cmd += ' --tmp_dir %s' % tmpdir
        cmd += ' --output_file %s' % profile_out
        cmd += ' --nproc %s' % params['cpus']
        cmd += ' --sample_id_key %s' % sam
        self.outputs['cmds'].setdefault((tech,), []).append(cmd)
        self.soft.add_status(tech, self.sam_pool, 1)
    else:
        io_update(self, i_f=bowtie2out, key=tech)
        self.soft.add_status(tech, self.sam_pool, 0)
    return sam_out, profile_out


def metaphlan(self) -> None:
    """MetaPhlAn is a computational tool for profiling the composition of
    microbial communities (Bacteria, Archaea and Eukaryotes) from metagenomic
    shotgun sequencing data (i.e. not 16S) with species-level. With the newly
    added StrainPhlAn module, it is now possible to perform accurate
    strain-level microbial profiling.
    MetaPhlAn relies on ~1.1M unique clade-specific marker genes (the latest
    marker information file mpa_v296_CHOCOPhlAn_201901_marker_info.txt.bz2
    can be found here) identified from ~100,000 reference genomes (~99,
    500 bacterial and archaeal and ~500 eukaryotic).
    What's new in version 3:
    - New MetaPhlAn marker genes extracted with a newer version of ChocoPhlAn
      based on UniRef
    - Estimation of metagenome composed by unknown microbes with parameter
      --unknown_estimation
    - Automatic retrieval and installation of the latest MetaPhlAn database
      with parameter --index latest
    - Virus profiling with --add_viruses
    - Calculation of metagenome size for improved estimation of reads mapped
      to a given clade
    - Inclusion of NCBI taxonomy ID in the ouput file
    - CAMI (Taxonomic) Profiling Output Format included
    - Removal of reads with low MAPQ values

    References
    ----------
    Beghini, Francesco, et al. "Integrating taxonomic, functional,
    and strain-level profiling of diverse microbial communities with
    bioBakery 3." Elife 10 (2021): e65088.

    Notes
    -----
    GitHub  : https://github.com/biobakery/MetaPhlAn/tree/3.0/
    Docs    : http://segatalab.cibio.unitn.it/tools/metaphlan/index.html
    Paper   : https://doi.org/10.7554/eLife.65088

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for humann
        .sam : str
            Sample name
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters for humann
        .databases
            All databases
        .config
            Configurations
    """
    for (tech, sam), inputs in self.inputs[self.sam_pool].items():
        if tech_specificity(self, inputs, tech, sam, ['illumina']):
            continue

        tmpdir = '$TMPDIR/metaphlan_%s_%s' % (tech, self.sam_pool)
        bowtie2out = '%s/%s/bowtie2/%s.bowtie2.bz2' % (self.dir, tech, sam)

        outs = profiling(self, tech, sam, inputs, bowtie2out, tmpdir)
        self.outputs['outs'].setdefault(
            (tech, sam), []).extend([bowtie2out] + list(outs))
        self.outputs['dir'] = [
            dirname(x) for x in self.outputs['outs'][(tech, sam)]]

        reads = get_read_count(self, tech, sam)
        if reads:
            analyses(self, tech, sam, bowtie2out, reads)
        if self.outputs['cmds'].get((tech,)):
            _cmd = ['mkdir -p %s\n' % tmpdir]
            self.outputs['cmds'][(tech,)] = _cmd + self.outputs['cmds'][(tech,)]


def get_profile(self) -> list:
    """Get the taxonomic profiles if there are some, for humann to only run
    of the taxa from these profiles.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for humann
        .sam : str
            Sample name
        .outputs : dict
            All outputs
        .soft.params
            Parameters for humann
        .databases
            All databases

    Returns
    -------
    profiles_dbs_outs : list
        Element are lists:
            profiles : list
                Paths to the taxonomic profile of the current sample.
            dbs : list
                Paths to the metaphlan database.
            outs : list
                Paths to the output folder.
    """
    profiles, dbs, outs = [], [], []
    db_path = self.databases.paths['humann']
    params = self.soft.params
    if params['profiles']:
        for profile_name, profile_fp in params['profiles'].items():
            # `profile_fp` must be "path/to/<sample>/profile_file.txt"
            profile_sams = {x.split('/')[-2]: x for x in glob.glob(profile_fp)}
            if self.sam_pool in profile_sams:
                profiles.append(profile_sams[self.sam_pool])
                io_update(self, i_f=profile_sams[self.sam_pool])
                dbs.append('')
                outs.append(
                    self.dir + '/profile_%s_%s_id%s_e%s_qC%s_sC%s_bypss%s' % (
                        profile_name,
                        params['uniref'],
                        params['id_thresh'],
                        params['evalue'],
                        params['query_coverage'],
                        params['subject_coverage'],
                        params['skip_translated']))
    else:
        profiles.append('')
        dbs.append(db_path)
        outs.append(self.dir + '/no_profile')
    profiles_dbs_outs = [profiles, dbs, outs]
    return profiles_dbs_outs


def get_cmd(
        self,
        profile: str,
        db: str,
        out: str,
        base: str,
        ali: str
) -> str:
    """Get the humann command line.

    Parameters
    ----------
    self : Commands class instance
        .sam : str
            Sample name
        .inputs : dict
            Input files
        .soft.params
            Parameters for humann
    profile : str
        Path to the taxonomic profile of the current sample.
    db : str
        Path to the metaphlan database.
    out : str
        Path to the output folder.
    base : str
        Path to fastq file.
    ali : str
        Path to temporary, alignment sam file

    Returns
    -------
    cmd : str
        humann command line.
    """
    cmd = 'cat %s > %s\n' % (' '.join(self.inputs[self.sam_pool]), base)
    cmd += 'humann'
    cmd += ' --input %s' % base
    cmd += ' --output %s' % out
    cmd += ' --prescreen-threshold %s' % self.soft.params['prescreen_threshold']
    cmd += ' --threads %s' % self.soft.params['cpus']
    cmd += ' --search-mode %s' % self.soft.params['uniref']
    if len(db):
        cmd += ' --metaphlan-options '
        cmd += '" --mpa_pkl %s/mpa_v20_m200.pkl' % db
    cmd += ' --bowtie2db %s"' % db
    cmd += ' --identity-threshold %s' % self.soft.params['identity']
    cmd += ' --evalue %s' % self.soft.params['evalue']
    cmd += ' --translated-query-coverage-threshold'
    cmd += ' %s' % self.soft.params['subject_cov']
    cmd += ' --translated-query-coverage-threshold'
    cmd += ' %s' % self.soft.params['subject_cov']
    if profile:
        cmd += ' --taxonomic-profile %s' % profile
    if self.soft.params['skip_translated']:
        cmd += ' --bypass-translated-search'
    if not to_do(ali):
        cmd += ' -r'
    cmd += ' --nucleotide-database %s' % self.soft.params['nucleotide_db']
    cmd += ' --protein-database  %s\n' % self.soft.params['protein_db']
    return cmd


def renorm(
        input_fp: str,
        relab_fp: str
) -> str:
    """Normalize the reads counts into proportions.

    Parameters
    ----------
    input_fp : str
        Path to the input file.
    relab_fp : str
        Path to the output file.

    Returns
    -------
    cmd : str
        Normalization command line.
    """
    cmd = 'humann2_renorm_table'
    cmd += ' --input %s' % input_fp
    cmd += ' --output %s' % relab_fp
    cmd += ' -u relab\n'
    return cmd


def get_ali(
        self,
        o: str
) -> str:
    """

    Parameters
    ----------
    self : Commands class instance
        .sam : str
            Sample name
        .outputs : dict
            All outputs
    o : str


    Returns
    -------
    ali : str
        Path to the alignment .sam file
    """
    ali = '%s/%s_tmp/%s_bowtie2_aligned.sam' % (o, self.sam_pool, self.sam_pool)
    if not self.config.force and not to_do(ali):
        io_update(self, i_f=ali)
    else:
        io_update(self, o_f=ali, o_d=o)
    return ali


def get_outputs(
        self,
        profile: str,
        db: str,
        out: str,
        ali: str,
        key: tuple
) -> tuple:
    """Get the profiling output files at genus, pathway and coverage levels.

    Parameters
    ----------
    self : Commands class instance
        .sam : str
            Sample name
        .outputs : dict
            All outputs
        .config
            Configurations
    profile : str
        Path to the taxonomic profile of the current sample.
    db : str
        Path to the metaphlan database.
    out : str
        Path to the output folder.
    ali : str
        Path to temporary, alignment sam file
    key : tuple
        Technology and/or smaple name

    Returns
    -------
    gen : str
        Path to the genus abundance profile output file
    pwy : str
        Path to the pathway abundance profile output file
    cov : str
        Path to the pathway coverage profile output file
    """
    base = '%s/%s.fastq.gz' % (out, self.sam_pool)
    gen = '%s/%s_genefamilies.tsv' % (out, self.sam_pool)
    pwy = '%s/%s_pathabundance.tsv' % (out, self.sam_pool)
    cov = '%s/%s_pathcoverage.tsv' % (out, self.sam_pool)
    if not self.config.force and not sum([to_do(x) for x in [gen, pwy, cov]]):
        io_update(self, i_f=[gen, pwy])
    else:
        cmd = get_cmd(self, profile, db, out, base, ali)
        self.outputs['cmds'].setdefault(key, []).append(cmd)
        io_update(self, o_f=[gen, pwy, cov])
    return gen, pwy, cov


def humann(self) -> None:
    """.

    References
    ----------


    Notes
    -----
    GitHub  :
    Docs    :
    Paper   :

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for humann
        .sam : str
            Sample name
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters for humann
        .databases
            All databases
        .config
            Configurations
    """
    for profile, db, out in zip(*get_profile(self)):
        key = (tech, '_'.join(profile, db))
        ali = get_ali(self, out)
        gen, pwy, cov = get_outputs(self, ali, profile, db, out, key)
        gen_relab = '%s/%s_genefamilies_relab.tsv' % (out, self.sam_pool)
        pwy_relab = '%s/%s_pathabundance_relab.tsv' % (out, self.sam_pool)
        if self.config.force or to_do(gen_relab) or to_do(pwy_relab):
            cmd_gen = renorm(gen, gen_relab)
            cmd_pwy = renorm(pwy, pwy_relab)
            self.outputs['cmds'].setdefault(key, []).append(cmd_gen)
            self.outputs['cmds'].setdefault(key, []).append(cmd_pwy)
            io_update(self, o_f=[gen_relab, pwy_relab], key=key)
        self.outputs['dirs'].append(out)
        self.outputs['outs'].append([out, ali, gen, gen_relab, pwy,
                                     pwy_relab, cov])


def get_sample_to_marker(
        self,
        tech,
        sam_dir: str,
        markers_dir: str
) -> None:
    """

    Parameters
    ----------
    self : Commands class instance
        .outputs : dict
            All outputs
        .config
            Configurations
    tech : str
    sam_dir : str
    markers_dir : str
    """

    sam_dir_ = sam_dir.replace('${SCRATCH_FOLDER}', '')
    to_dos = status_update(self, tech, [sam_dir], folder=True)
    if self.config.force or len(glob.glob('%s/*.sam.bz2' % sam_dir_)):
        cmd = 'sample2markers.py'
        cmd += ' -i %s/*.sam.bz2' % sam_dir
        cmd += ' -o %s' % markers_dir
        cmd += ' -n %s' % self.soft.params['cpus']
        if to_dos:
            self.outputs['cmds'].setdefault((tech,), []).append(False)
        else:
            self.outputs['cmds'].setdefault((tech,), []).append(cmd)


def extract_markers(
        self,
        tech,
        key,
        st: str,
        strain_dir: str,
        db_dir: str
) -> None:
    """Extract markers for the current species strains

    Parameters
    ----------
    self : Commands class instance
        .config
            Configurations
    tech
    key
    st : str
    strain_dir : str
    db_dir : str
    """
    to_dos = status_update(self, tech, [strain_dir], folder=True)
    if self.config.force or to_do('%s/%s.fna' % (strain_dir, st)):
        cmd = 'extract_markers.py'
        cmd += ' -c %s' % st
        cmd += ' -o %s' % db_dir
        if to_dos:
            self.outputs['cmds'].setdefault(key, []).append(False)
        else:
            self.outputs['cmds'].setdefault(key, []).append(cmd)


def strainphlan(self) -> None:
    """StrainPhlAn is a computational tool for tracking individual strains
    across a large set of samples. The input of StrainPhlAn is a set of
    metagenomic samples and for each species, the output is a multiple
    sequence alignment (MSA) file of all species strains reconstructed
    directly from the samples. From this MSA, StrainPhlAn calls (PhyloPhlAn
    3)[http://segatalab.cibio.unitn.it/tools/phylophlan3/index.html] to build
    the phylogenetic tree showing the strain evolution of the sample strains.

    References
    ----------
    Beghini, Francesco, et al. "Integrating taxonomic, functional,
    and strain-level profiling of diverse microbial communities with
    bioBakery 3." Elife 10 (2021): e65088.

    Notes
    -----
    Docs    : https://github.com/biobakery/MetaPhlAn/wiki/StrainPhlAn-3
    Website : http://segatalab.cibio.unitn.it/tools/phylophlan3/index.html
    Paper   : https://doi.org/10.7554/eLife.65088

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for humann
        .sam : str
            Sample name
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters for humann
        .databases
            All databases
        .config
            Configurations
    """
    for (tech, sam), inputs in self.inputs[self.sam_pool].items():
        if tech_specificity(self, inputs, 'illumina', sam):
            continue
        bt2_fp, sam_fp = inputs[:2]
        sam_dir = dirname(sam_fp)
        db_dir = '%s/db_markers' % self.dir
        meta_dir = '%s/metadata' % self.dir
        markers_dir = '%s/consensus_markers' % self.dir
        wol_pd = self.databases.paths['wol']['taxonomy']

        self.outputs['outs'] = dict({})
        self.outputs['dirs'].extend([db_dir, meta_dir])
        get_sample_to_marker(self, tech, sam_dir, markers_dir)

        for strain_group, strains in self.config.strains.items():

            tmp = '$TMPDIR/strainphlan_%s' % strain_group
            key = (tech, strain_group)
            io_update(self, i_d=sam_dir, o_d=[markers_dir, db_dir, meta_dir],
                      key=key)

            strain_dir = '%s/%s' % (db_dir, strain_group)
            self.outputs['dirs'].append(strain_dir)
            for strain in [x.replace(' ', '_') for x in strains]:
                if strain[0] != 's':
                    continue
                extract_markers(self, tech, key, strain, strain_dir, db_dir)
                odir = '%s/output/%s' % (self.dir, strain_group)
                self.outputs['dirs'].append(odir)
                self.outputs['outs'][(strain_group, strain)] = odir
                io_update(self, o_d=odir, key=key)
                tree = '%s/RAxML_bestTree.%s.StrainPhlAn3.tre' % (odir, strain)
                to_dos = status_update(self, tech, [markers_dir, db_dir],
                                       folder=True)
                if self.config.force or to_do(tree):
                    cmd = 'strainphlan'
                    cmd += ' -s %s/*.pkl' % markers_dir
                    cmd += ' -m %s/%s.fna' % (db_dir, strain)
                    if strain in wol_pd['species']:
                        fna = '%s/%s.fna.bz2' % (
                            self.databases.paths['wol']['fna'],
                            wol_pd.loc[wol_pd['species'] == strain, 0])
                        io_update(self, i_f=fna, key=key)
                        cmd += ' -r %s' % fna
                    cmd += ' -o %s' % odir
                    cmd += ' -n %s' % self.soft.params['cpus']
                    cmd += ' -c %s' % strain
                    cmd += ' --mutation_rates'
                    cmd = 'mkdir -p %s\n' % tmp + cmd
                    if to_dos:
                        self.outputs['cmds'].setdefault(key, []).append(False)
                    else:
                        self.outputs['cmds'].setdefault(key, []).append(cmd)
                    self.soft.add_status(
                        tech, sam, 1, group=strain_group, genome=strain)
                else:
                    self.soft.add_status(
                        tech, sam, 0, group=strain_group, genome=strain)



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

from metagenomix._io_utils import caller, io_update


def bwa():
    cmd = 'bwa index -p %s/index %s\n' % (out_dir, contigs)
    cmd += 'bwa mem -p %s/index %s' % (out_dir, fastqs)
    cmd += ' | samtools view -bh'
    cmd += ' | samtools sort -o %s/%s\n' % (out_dir, bam)
    cmd += 'samtools index %s %s\n' % (bam, bai)


def bowtie2():
    cmd = 'bwa index -p %s/index %s\n' % (out_dir, contigs)
    cmd += 'bwa mem -p %s/index %s' % (out_dir, fastqs)
    cmd += ' | samtools view -bh'
    cmd += ' | samtools sort -o %s/%s\n' % (out_dir, bam)
    cmd += 'samtools index %s %s\n' % (bam, bai)


def spades(self):
    pass


def metawrap_refine(self):
    pass


def mapping(self):
    """Mapping would be a rather specific process that generically consists
    in aligning reads that can be raw, filtered, or for a specific scope (
    e.g., that were used to re-assemble a MAGs or specific proteins).

    It can be done using different aligners, such as BWA, bowtie2, minimap2,
    bbmap, or others, but all should yield a .sam file that can be piped in
    to samtools in order to obtain one .bam file and one .bai file per
    mapping. These files are then possibly used by several other softwares.

    Parameters
    ----------
    self : Commands class instance
        Contains all the attributes needed for binning on the current sample
    """
    # This function splits the name of the software and calls as function
    # the last underscore-separated field (which is in this module)
    module_call = caller(self, __name__)
    module_call(self)


def previous_mapper(self):
    refs = {}
    fastqs, group_fps = self.config.fastq, self.inputs[self.pool]
    if self.soft.prev == 'drep':
        for algo, fp in self.inputs[self.pool].items():
            ref_fasta = '%s.fa' % fp
            if not isfile(ref_fasta):
                cmd = 'if [ ! -f %s ]; then cat %s/*.fa > %s; fi' % (
                    ref_fasta, fp, ref_fasta)
                self.outputs['cmds'].setdefault(
                    (self.pool, algo), []).append(cmd)
            refs[algo] = ref_fasta
    elif self.soft.prev == 'metawrap_reassemble':
        # needs revision...
        refs = {'': glob.glob(self.inputs[self.pool][self.sam][1])}
    elif self.soft.prev == 'spades':
        refs = {group: fps[1] for group, fps in group_fps.items()}

    self.outputs['outs'] = {}
    for group, fasta in refs.items():
        self.outputs['outs'][group] = {}
        out_dir = '%s/%s' % (self.dir, self.pool)
        if group:
            out_dir += '/%s' % group
        self.outputs['dirs'].append(out_dir)
        for sam in self.pools[self.pool][group]:
            bam_out = '%s/%s.bam' % (out_dir, sam)
            self.outputs['outs'][group][sam] = bam_out
            cmd = 'minimap2'
            cmd += ' -a -x sr'
            cmd += ' %s' % fasta
            cmd += ' %s' % fastqs[sam][0]
            cmd += ' %s' % fastqs[sam][1]
            cmd += ' | samtools view -F 0x104 -b -'
            cmd += ' | samtools sort -o %s - ' % bam_out
            if not isfile(bam_out) or self.config.force:
                self.outputs['cmds'].setdefault(
                    (self.pool, group), []).append(cmd)
            bam_out_bai = '%s.bai' % bam_out
            cmd = 'samtools index %s' % bam_out
            if not isfile(bam_out_bai) or self.config.force:
                self.outputs['cmds'].setdefault(
                    (self.pool, group), []).append(cmd)
            io_update(self, i_f=([fasta] + fastqs[sam]),
                      o_f=[bam_out, bam_out_bai])


def prep_map__spades_prodigal(self):
    if 'prodigal' not in self.softs or 'mapping' not in self.softs:
        return None
    if self.softs['prodigal'].prev != 'spades':
        return None
    if self.softs['mapping'].prev != 'spades':
        return None
    prodigals_fps = self.softs['prodigal'].outputs
    sams_fps = self.softs['mapping'].outputs
    group_fps = self.inputs[self.pool]
    self.outputs['outs'] = {}
    for group, fps in group_fps.items():
        self.outputs['outs'][group] = {}
        for sam in self.pools[self.pool][group]:
            bam = sams_fps[self.pool][group][sam]
            prot = prodigals_fps[self.pool][group][1]
            out_dir = '%s/%s/%s' % (self.dir, self.pool, sam)
            out = '%s/reads.txt' % out_dir
            if not isfile(out):
                cmd = 'pysam_reads_to_prodigal.py \\\n'
                cmd += '-prodigal %s \\\n' % prot
                cmd += '-bam %s \\\n' % bam
                cmd += '-out %s\n' % out
                self.outputs['cmds'].setdefault(
                    (self.pool, group), []).append(cmd)
            self.outputs['outs'][group][sam] = out
            io_update(self, i_f=[prot, bam, out])


def map__cazy_spades(self):
    pass


def map__cazy_macsyfinder(self):
    pass


def map__spades_bins(self):
    pass


def map__drep(self):
    pass




#!/usr/bin/env python3
"""Generate a list of download links to genome sequences.

Usage:
    bzcat metadata.tsv.xz | make_down_list.py > downlist.txt
"""

import fileinput


def main():
    c1, c2 = 'assembly_accession', 'asm_name'
    i1, i2 = 0, 0
    header = True
    for line in fileinput.input():
        x = line.rstrip('\r\n').split('\t')
        if header:
            try:
                i1, i2 = x.index(c1), x.index(c2)
            except ValueError:
                raise ValueError(
                    'Column "%s" or "%s" is not found.' % (c1, c2))
            header = False
        else:
            print(get_fna_path(x[i1], x[i2]))


def get_fna_path(asb, asm):
    asm = asm.replace(' ', '_').replace('#', '_')
    srv = 'ftp://ftp.ncbi.nlm.nih.gov/genomes/all'
    ext = 'genomic.fna.gz'
    return ('%s/%s/%s/%s/%s/%s_%s/%s_%s_%s'
            % (srv, asb[:3], asb[4:7], asb[7:10], asb[10:13],
               asb, asm, asb, asm, ext))


if __name__ == '__main__':
    main()
